'use strict';

var react = require('react');

const defaultReturn = {
    isValid: false,
    value: null,
};
function getRadioValue(options) {
    return Array.isArray(options)
        ? options.reduce((previous, { ref: { name, checked, value } }) => checked
            ? {
                isValid: true,
                value,
            }
            : previous, defaultReturn)
        : defaultReturn;
}

function getMultipleSelectValue(options) {
    return options.filter(({ selected }) => selected).map(({ value }) => value);
}

function isRadioInput(type) {
    return type === 'radio';
}

function getFieldValue(fields, { type, name, options, checked, value }) {
    if (isRadioInput(type)) {
        return getRadioValue(fields[name].options).value;
    }
    else if (type === 'select-multiple') {
        return getMultipleSelectValue(options);
    }
    else if (type === 'checkbox') {
        return checked;
    }
    return value;
}

function getFieldsValues(fields, filedNames) {
    return Object.values(fields).reduce((previous, { ref, ref: { name } }) => {
        const value = getFieldValue(fields, ref);
        if (typeof filedNames === 'string') {
            if (name === filedNames) {
                return value;
            }
            return previous;
        }
        const copy = Object.assign({}, (previous || {}));
        if (Array.isArray(filedNames)) {
            if (filedNames.includes(name)) {
                copy[name] = value;
            }
        }
        else {
            copy[name] = value;
        }
        return copy;
    }, undefined);
}

const DATE_INPUTS = ['date', 'time', 'month', 'datetime', 'datetime-local', 'week'];
const STRING_INPUTS = ['text', 'email', 'password', 'search', 'tel', 'url'];

var validateField = ({ ref: { type, value, name, checked }, required, maxLength, minLength, min, max, pattern, validate }, fields) => {
    const copy = {};
    if (required &&
        ((type === 'checkbox' && !checked) ||
            (!isRadioInput(type) && type !== 'checkbox' && value === '') ||
            (isRadioInput(type) && !getRadioValue(fields[name].options).isValid))) {
        copy[name] = {
            required: true,
        };
    }
    if (value === '')
        return copy;
    // min and max section
    if (min || max) {
        let exceedMax;
        let exceedMin;
        const valueNumber = parseFloat(value);
        if (type === 'number') {
            exceedMax = max && valueNumber > max;
            exceedMin = min && valueNumber < min;
        }
        else if (DATE_INPUTS.includes(type)) {
            exceedMax = max && new Date(value) > new Date(max);
            exceedMin = min && new Date(value) < new Date(min);
        }
        if (exceedMax || exceedMin) {
            copy[name] = Object.assign({}, copy[name], (exceedMax ? { max: true } : null), (exceedMin ? { min: true } : null));
        }
    }
    if ((maxLength || minLength) && STRING_INPUTS.includes(type)) {
        const exceedMax = maxLength && value.toString().length > maxLength;
        const exceedMin = minLength && value.toString().length < minLength;
        if (exceedMax || exceedMin) {
            copy[name] = Object.assign({}, copy[name], (exceedMax ? { maxLength: true } : null), (exceedMin ? { minLength: true } : null));
        }
    }
    if (pattern && pattern instanceof RegExp && !pattern.test(value)) {
        copy[name] = Object.assign({}, copy[name], { pattern: true });
    }
    if (validate) {
        if (typeof validate === 'function') {
            if (!validate(value)) {
                copy[name] = Object.assign({}, copy[name], { validate: true });
            }
        }
        else if (typeof validate === 'object') {
            copy[name] = Object.assign({}, copy[name], { validate: Object.entries(validate).reduce((previous, [key, validate]) => {
                    if (typeof validate === 'function' && !validate(value)) {
                        previous[key] = true;
                    }
                    return previous;
                }, {}) });
        }
    }
    return copy;
};

var removeAllEventListeners = (ref, validateWithStateUpdate) => {
    ref.removeEventListener('input', validateWithStateUpdate);
    ref.removeEventListener('change', validateWithStateUpdate);
    ref.removeEventListener('blur', validateWithStateUpdate);
};

function findMissDomAndClean({ target: { ref, ref: { name, type }, mutationWatcher, options, }, fields, validateWithStateUpdate, forceDelete = false, }) {
    if (isRadioInput(type) && options) {
        options.forEach(({ ref }, index) => {
            if (!document.body.contains(ref) && fields[name] && fields[name].options && fields[name].options[index]) {
                removeAllEventListeners(fields[name].options[index], validateWithStateUpdate);
                fields[name].options[index].mutationWatcher.disconnect();
                fields[name].options.splice(index, 1);
            }
        });
        if (!fields[name].options.length) {
            delete fields[name];
            return fields;
        }
    }
    else if (ref && (!document.body.contains(ref) || forceDelete)) {
        removeAllEventListeners(ref, validateWithStateUpdate);
        if (mutationWatcher)
            mutationWatcher.disconnect();
        delete fields[name];
        return fields;
    }
    return fields;
}

function onDomRemove(element, onDetachCallback) {
    const observer = new MutationObserver(() => {
        function isDetached(el) {
            if (el.parentNode === window.document) {
                return false;
            }
            else if (el.parentNode === null) {
                return true;
            }
            return isDetached(el.parentNode);
        }
        if (isDetached(element)) {
            observer.disconnect();
            onDetachCallback();
        }
    });
    observer.observe(window.document, {
        childList: true,
        subtree: true,
    });
    return observer;
}

function attachEventListeners({ mode, allFields, watchFields, radioOptionIndex, ref, type, name, validateWithStateUpdate, }) {
    const field = allFields[name];
    const isOnChange = mode === 'onChange' || watchFields.current[ref.name];
    const isOnBlur = mode === 'onBlur';
    if (!field || (!isOnChange && !isOnBlur))
        return;
    const isRadio = isRadioInput(type);
    if (isRadio) {
        const options = field.options;
        if (!options[radioOptionIndex])
            return;
        const eventAttached = options[radioOptionIndex].eventAttached;
        if (eventAttached) {
            removeAllEventListeners(ref, validateWithStateUpdate);
        }
        else {
            options[radioOptionIndex].ref.addEventListener(isOnChange ? 'change' : 'blur', validateWithStateUpdate);
        }
        options[radioOptionIndex].eventAttached = !eventAttached;
    }
    else {
        const eventAttached = field.eventAttached;
        if (eventAttached) {
            removeAllEventListeners(ref, validateWithStateUpdate);
        }
        else {
            ref.addEventListener(isOnChange ? 'input' : 'blur', validateWithStateUpdate);
        }
        field.eventAttached = !eventAttached;
    }
}

function useForm({ mode } = { mode: 'onSubmit' }) {
    const fields = react.useRef({});
    const localErrorMessages = react.useRef({});
    const watchFields = react.useRef({});
    const [errors, updateErrorMessage] = react.useState({});
    function validateWithStateUpdate({ target: { name }, type }) {
        const ref = fields.current[name];
        const error = validateField(ref, fields.current);
        if (localErrorMessages.current[name] !== error[name] ||
            mode === 'onChange' ||
            (mode === 'onBlur' && type === 'blur') ||
            watchFields.current[name]) {
            const copy = Object.assign({}, localErrorMessages.current, error);
            if (!error[name])
                delete copy[name];
            localErrorMessages.current = copy;
            updateErrorMessage(copy);
        }
    }
    function removeReferenceAndEventListeners(data, forceDelete = false) {
        findMissDomAndClean({
            target: data,
            fields: fields.current,
            validateWithStateUpdate,
            forceDelete,
        });
    }
    function register(data) {
        if (!data || !data.ref)
            return;
        if (!data.ref.name) {
            console.warn('Oops missing the name for field:', data.ref);
            return;
        }
        let radioOptionIndex;
        const { ref, required, ref: { name, type, value }, } = data;
        const allFields = fields.current;
        if (isRadioInput(type)) {
            if (!allFields[name]) {
                allFields[name] = { options: [], required, ref: { type: 'radio', name } };
            }
            const options = allFields[name].options || [];
            radioOptionIndex = options.findIndex(({ ref }) => value === ref.value);
            if (radioOptionIndex > -1) {
                options[radioOptionIndex] = Object.assign({}, options[radioOptionIndex], data);
            }
            else {
                options.push(Object.assign({}, data, { mutationWatcher: onDomRemove(ref, () => removeReferenceAndEventListeners(data, true)) }));
                radioOptionIndex = options.length - 1;
            }
        }
        else {
            allFields[name] = data;
            if (!allFields[name]) {
                allFields[name].mutationWatcher = onDomRemove(ref, () => removeReferenceAndEventListeners(data, true));
            }
        }
        attachEventListeners({ allFields, watchFields, ref, type, radioOptionIndex, name, mode, validateWithStateUpdate });
    }
    function watch(filedNames, defaultValue) {
        if (typeof filedNames === 'string') {
            if (!watchFields.current[filedNames])
                watchFields.current[filedNames] = true;
        }
        else if (Array.isArray(filedNames)) {
            filedNames.forEach(name => {
                if (!watchFields.current[name])
                    return;
                watchFields.current[name] = true;
            });
        }
        else {
            Object.values(fields.current).forEach(({ ref }) => {
                if (!ref)
                    return;
                if (!watchFields.current[ref.name])
                    return;
                watchFields.current[ref.name] = true;
            });
        }
        const result = getFieldsValues(fields.current, filedNames);
        return result === undefined ? defaultValue : result;
    }
    const handleSubmit = (callback) => e => {
        e.preventDefault();
        const allFields = fields.current;
        const { localErrors, values } = Object.values(allFields).reduce((previous, data) => {
            const { ref, ref: { name, type }, options, } = data;
            removeReferenceAndEventListeners(data);
            if (!fields.current[name])
                return previous;
            const fieldError = validateField(data, allFields);
            const hasError = fieldError && fieldError[name];
            if (!hasError) {
                previous.values[name] = getFieldValue(allFields, ref);
                return previous;
            }
            if (!fields.current[name].eventAttached) {
                if (isRadioInput(type) && Array.isArray(options)) {
                    options.forEach(option => {
                        if (option.eventAttached)
                            return;
                        option.ref.addEventListener('change', validateWithStateUpdate);
                        option.eventAttached = true;
                    });
                }
                else {
                    ref.addEventListener('input', validateWithStateUpdate);
                    data.eventAttached = true;
                }
            }
            previous.localErrors = Object.assign({}, previous.localErrors, fieldError);
            return previous;
        }, {
            localErrors: {},
            values: {},
        });
        if (JSON.stringify(localErrorMessages.current) !== JSON.stringify(localErrors)) {
            updateErrorMessage(localErrors);
            localErrorMessages.current = localErrors;
        }
        if (!Object.values(localErrors).length)
            callback(values, e);
    };
    react.useEffect(() => () => {
        fields.current &&
            Object.values(fields.current).forEach(({ ref, options }) => isRadioInput(ref.type) && Array.isArray(options)
                ? options.forEach(({ ref }) => removeAllEventListeners(ref, validateWithStateUpdate))
                : removeAllEventListeners(ref, validateWithStateUpdate));
        fields.current = {};
        watchFields.current = {};
        localErrorMessages.current = {};
        updateErrorMessage({});
    }, [mode]);
    return {
        register,
        handleSubmit,
        errors,
        watch,
    };
}

module.exports = useForm;
